using System;
using System.IO;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace FTPDemo
{
	/// <summary>
	/// Summary description for clsLocation.
	/// </summary>
	public class clsFTP
	{
		// Error codes
		public const int FTP_NO_ERROR = 0;
		public const int FTP_TOO_MANY_INTERNET_HANDLES = 1;
		public const int FTP_TIMEOUT = 2;
		public const int FTP_INTERNAL_ERROR = 3;
		public const int FTP_INVALID_URL = 4;
		public const int FTP_NAME_NOT_RESOLVED = 5;
		public const int FTP_PROTOCOL_NOT_FOUND = 6;
		public const int FTP_SHUTDOWN = 7;
		public const int FTP_INCORRECT_USER_NAME = 8;
		public const int FTP_INCORRECT_PASSWORD = 9;
		public const int FTP_LOGIN_FAILED = 10;
		public const int FTP_REQUEST_PENDING = 11;
		public const int FTP_CANNOT_CONNECT = 12;
		public const int FTP_CONNECTION_ABORTED = 13;
		public const int FTP_CONNECTION_RESET = 14;
		public const int FTP_FORCE_RETRY = 15;
		public const int FTP_INVALID_PROXY_REQUEST = 16;
		public const int FTP_TRANSFER_IN_PROGRESS = 17;
		public const int FTP_DROPPED = 18;
		public const int FTP_NO_PASSIVE_MODE = 19;
		public const int FTP_LOCAL_FILE_EXISTS = 20;
		public const int FTP_UNKNOWN = 21;

		// Overwrite modes
		public const int FTP_OVERWRITE = 1;
		public const int FTP_DONT_OVERWRITE = 0;

		// Transfer modes
		public const int FTP_TRANSFER_UNKNOWN = 0;
		public const int FTP_TRANSFER_ASCII = 1;
		public const int FTP_TRANSFER_BINARY = 2;

		public const int DEFAULT_FTP_PORT = 21;

		// FTP Settings
		public static string sFTPSession;
		public static string sFTPUsername;
		public static string sFTPPassword;
		public static string sURL;
		public static string sLocalFilter;
		public static string sRemoteFilter;

		public static string [] sSessions = new string[3];
		public static string [] sUsernames = new string[3];
		public static string [] sPasswords = new string[3];
		public static int iPassive;
		public static string sLocalPath;

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPOpen(string sServer, int iPort, string sUser, string sPwd, int iPassive, int iTimeOut, int iLicence);
		
		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPGet(string sRemoteFile, string sLocalFile, byte bOverWriteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPull(string sRemoteFile, string sLocalFile, byte bOverWriteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPullStart(string sRemoteFile, string sLocalFile, byte bOverWriteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPullMore();

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPullEnd();

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPut(string sLocalFile, string sRemoteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPush(string sLocalFile, string sRemoteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPushStart(string sLocalFile, string sRemoteFile, int iTransferType, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPushMore();

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPPushEnd();

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPClose(int iLicense);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPChangeDirectory(string sPath, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPCurrentDirectory(string sCurrentDirectory, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPFileExists(string sFile, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPFindFirstFile(string sFileString, string sFilter, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPFindNextFile(string sFileString, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPRemoveFile(string sRemoteFile, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPRenameFile(string sOldFile, string sNewFile, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPRemoveDirectory(string sDirectory, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPMakeDirectory(string sDirectory, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPUploadDirectory(string sLocalDirectory, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPDownloadDirectory(string sRemoteDirectory, string sLocalPath, int iLicence);

		[DllImport("SapphireFTP.dll")]
		public static extern void sFTPErrorText(string sErrorText, int iError);

		[DllImport("SapphireFTP.dll")]
		public static extern int sFTPReconnect(int iPassive, int iLicence);


		public clsFTP()
		{
		}

		public static void InitialiseFileManager()
		{
			sLocalPath = "\\";
		}

		public static void SetFTPSettings()
		{
			clsListItem [] oComboSession = new clsListItem[3];
			clsMain.fSetFTP.cboSession.Items.Clear();
			clsMain.fSetFTP.cboSession.DisplayMember = "sText";
			clsMain.fSetFTP.cboSession.ValueMember = "iValue";

			// Set up the default settings for FTP Web site
			sSessions[1] = "ftp.sapphire-solutions.co.uk";
			sUsernames[1] = "sapphirepublic";
			sPasswords[1] = "solutions";
			oComboSession[1] = new clsListItem();
			oComboSession[1].iValue = 1;
			oComboSession[1].sText = sSessions[1];
			clsMain.fSetFTP.cboSession.Items.Add(oComboSession[1]);

			sSessions[2] = "pentium4";
			sUsernames[2] = "anonymous";
			sPasswords[2] = "sales@sapphire-solutions.co.uk";
			oComboSession[2] = new clsListItem();
			oComboSession[2].iValue = 2;
			oComboSession[2].sText = sSessions[2];
			clsMain.fSetFTP.cboSession.Items.Add(oComboSession[2]);

			// Make the first one the default
			clsMain.fSetFTP.cboSession.SelectedIndex = 0;
			clsMain.fSetFTP.txtUsername.Text = sUsernames[1];
			clsMain.fSetFTP.txtPassword.Text = sPasswords[1];

			sFTPSession = sSessions[1];
			sFTPUsername = sUsernames[1];
			sFTPPassword = sPasswords[1];
		}

		public static int sLocalChangeDirectory(string sDirectory)
		{
			string sDir;
			int iCounter;
			int iReturn;

			iReturn = 1;

			if (sDirectory == "..")
			{
				if (sLocalPath != "\\")
				{
					// Remove the last folder in the string
					sDir = sLocalPath.Substring(0,sLocalPath.Length-1);
					iCounter = sDir.Length;
					while (sDir.Substring(iCounter-1, 1) != "\\" && iCounter > 0)
					{
						sDir = sDir.Substring(0,iCounter-1);
						iCounter--;
					}

					sLocalPath = sDir;
					iReturn = 0;
				}
			}
			else
			{
				sLocalPath = sLocalPath + sDirectory + "\\";
				iReturn = 0;
			}
			return iReturn;
		}


		public static string sLocalCurrentDirectory()
		{
			return sLocalPath;
		}

		public static void sLocalRefreshList()
		{
			ListViewItem liFile;
			DirectoryInfo fsDir = new DirectoryInfo(sLocalPath);
			FileInfo [] sFiles;
			DirectoryInfo [] sDirectories;
			int iCount;

			clsMain.fFTPExplorer.lstLocal.Items.Clear();

			clsMain.fFTPExplorer.lblLocal.Text = "Local Directory: " + sLocalPath + sLocalFilter;

			if (sLocalPath != "\\")
			{
				liFile = new ListViewItem("[..]");
				liFile.SubItems.Add("");
				liFile.SubItems.Add("");
				liFile.SubItems.Add("DIR");
				liFile = clsMain.fFTPExplorer.lstLocal.Items.Add(liFile);
			}

			// Get all the sub directories
			try // GetDirectories crashes if there are no contents in the folder
			{
				sDirectories = fsDir.GetDirectories();
				for (iCount = 0; iCount <= sDirectories.GetUpperBound(0); iCount++)
				{
					liFile = new ListViewItem("[" + sDirectories[iCount].Name + "]");
					// Get the size
					liFile.SubItems.Add("");
					// Get the date
					liFile.SubItems.Add(sDirectories[iCount].CreationTime.ToShortDateString() + " " + sDirectories[iCount].CreationTime.ToLongTimeString());
					// Get the attributes of the directory
					liFile.SubItems.Add(GetDirectoryAttributes(sDirectories[iCount]));
					clsMain.fFTPExplorer.lstLocal.Items.Add(liFile);
				}
			}
			catch
			{
			}

			try
			{
				sFiles = fsDir.GetFiles(sLocalFilter);
				for (iCount = 0; iCount <= sFiles.GetUpperBound(0); iCount++)
				{
					liFile = new ListViewItem(sFiles[iCount].Name);
					// Get the size
					liFile.SubItems.Add(GetFileSizeText(Convert.ToInt32(sFiles[iCount].Length)));
					// Get the date
					liFile.SubItems.Add(sFiles[iCount].CreationTime.ToShortDateString() + " " + sFiles[iCount].CreationTime.ToLongTimeString());
					// Get the attributes of the directory
					liFile.SubItems.Add(GetFileAttributes(sFiles[iCount]));
					clsMain.fFTPExplorer.lstLocal.Items.Add(liFile);
				}
			}
			catch
			{
			}
		}

		public static int sLocalRemoveFile(string sLocalFile)
		{
			FileInfo sFile = new FileInfo(sLocalPath + sLocalFile);
			int iReturn = 0;

			try
			{
				sFile.Delete();
			}
			catch
			{
				iReturn = 1;
			}
			return iReturn;
		}

		public static int sLocalRenameFile(string sOldFile, string sNewFile)
		{
			FileInfo sFile = new FileInfo(sLocalPath + sOldFile);
			int iReturn = 0;

			try
			{
				sFile.MoveTo(sLocalPath + sNewFile);
			}
			catch
			{
				iReturn = 1;
			}
			return iReturn;
		}


		public static int sLocalRemoveDirectory(string sDirectory)
		{
			DirectoryInfo fsDir = new DirectoryInfo(sLocalPath + sDirectory);
			FileInfo [] sFiles;
			DirectoryInfo [] sDirectories;
			int iCount;

			// Search for all sub directories and remove their content
			try // GetDirectories crashes if there are no contents in the folder
			{
				sDirectories = fsDir.GetDirectories();
				for (iCount = 0; iCount <= sDirectories.GetUpperBound(0); iCount++)
				{
					sLocalRemoveDirectory(sDirectory + "\\" + sDirectories[iCount].Name);
				}
			}
			catch
			{
				return 1;
			}

			// Remove all the files in this directory
			try
			{
				sFiles = fsDir.GetFiles();
				for (iCount = 0; iCount <= sFiles.GetUpperBound(0); iCount++)
				{
					sFiles[iCount].Delete();
				}
			}
			catch
			{
				return 1;
			}

			// Remove this directory
			fsDir.Delete(false);
			return 0;
		}

		public static int sLocalMakeDirectory(string sDirectory)
		{
			DirectoryInfo fsDir = new DirectoryInfo(sLocalPath);
			try
			{
				fsDir.CreateSubdirectory(sDirectory);
				return 0;
			}
			catch
			{
				return 1;
			}
		}

		public static string GetFileSizeText(int iSize)
		{
			if (iSize < 1024)
			{
				if (iSize == 1)
					return "1 byte";
				else
					return iSize.ToString() + " bytes";
			}
			else
			{
				if (iSize < 1024 * 1024)
					return Convert.ToString(iSize / 1024) + " KB";
				else
					return Convert.ToString(iSize / (1024 * 1024)) + " MB";
			}
		}
			
		private static string GetDirectoryAttributes(DirectoryInfo sDir)
		{
			string sAttributes;
			bool bComma;

			bComma = false;
			sAttributes = "";

			if (sDir.Attributes == FileAttributes.ReadOnly)
			{
				if (bComma  == true)
					sAttributes = ",RD";
				else
					sAttributes = "RD";
				bComma = true;
			}
			if (sDir.Attributes == FileAttributes.Hidden)
			{
				if (bComma == true)
					sAttributes = ",HID";
				else
					sAttributes = "HID";
				bComma = true;
			}
			if (sDir.Attributes == FileAttributes.System)
			{
				if (bComma == true)
					sAttributes = ",SYS";
				else
					sAttributes = "SYS";
				bComma = true;
			}

			if (bComma == true)
				sAttributes = ",DIR";
			else
				sAttributes = "DIR";
			bComma = true;

			if (sDir.Attributes == FileAttributes.Archive)
			{
				if (bComma == true)
					sAttributes = ",ARC";
				else
					sAttributes = "ARC";
				bComma = true;
			}

			return sAttributes;
		}

		private static string GetFileAttributes(FileInfo sFile)
		{
			string sAttributes;
			bool bComma;

			bComma = false;
			sAttributes = "";

			if (sFile.Attributes == FileAttributes.ReadOnly)
			{
				if (bComma == true)
					sAttributes = ",RD";
				else
					sAttributes = "RD";
				bComma = true;
			}
			if (sFile.Attributes == FileAttributes.Hidden)
			{
				if (bComma == true)
					sAttributes = ",HID";
				else
					sAttributes = "HID";
				bComma = true;
			}
			if (sFile.Attributes == FileAttributes.System)
			{
				if (bComma == true)
					sAttributes = ",SYS";
				else
					sAttributes = "SYS";
				bComma = true;
			}
			if (sFile.Attributes == FileAttributes.Archive)
			{
				if (bComma == true)
					sAttributes = ",ARC";
				else
					sAttributes = "ARC";
				bComma = true;
			}

			return sAttributes;
		}

		public static bool sLocalFileExists(string sLocalFile)
		{
			FileInfo sFile = new FileInfo(sLocalFile);
			return sFile.Exists;
		}

		public static string GetFTPError(int iError)
		{
			string sErrorText = new String(' ',1024);
	
			clsFTP.sFTPErrorText(sErrorText, iError);
		
			sErrorText = sErrorText.TrimEnd(' ');
			sErrorText = sErrorText.TrimEnd(Convert.ToChar(0x0000));
			return sErrorText;
		}			

		public static string FTPCurrentDirectory()
		{
			string sReturn = new String(' ',260);

			if (clsFTP.sFTPCurrentDirectory(sReturn, clsMain.SapphireFTPLicence) == 0)
				sReturn = "//";
		
			sReturn = sReturn.TrimEnd(' ');
			sReturn = sReturn.TrimEnd(Convert.ToChar(0x0000));
			return sReturn;
		}

		public static string FTPFindFirstFile(string sRemoteFilter)
		{
			string sFile = new String(' ',1024);
			if (sFTPFindFirstFile(sFile, sRemoteFilter, clsMain.SapphireFTPLicence) == 0)
				sFile = "";
			sFile = sFile.TrimEnd(' ');
			sFile = sFile.TrimEnd(Convert.ToChar(0x0000));
			return sFile;
		}
		
		public static string FTPFindNextFile()
		{
			string sFile = new String(' ',1024);
		
			if (sFTPFindNextFile(sFile, clsMain.SapphireFTPLicence) == 0)
				sFile = "";
	
			sFile = sFile.TrimEnd(' ');
			sFile = sFile.TrimEnd(Convert.ToChar(0x0000));
			return sFile;
		}
	}
}